library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.OptoUtils.all;


entity Bram is
	generic (
		width : natural := 64;
		adr_width : natural := 10
	 );
    Port (
				 clk : IN STD_LOGIC;
				 reset : IN STD_LOGIC;
				 we : IN STD_LOGIC;
				 adr_r : IN STD_LOGIC_VECTOR(adr_width-1 DOWNTO 0);
				 adr_w : IN STD_LOGIC_VECTOR(adr_width-1 DOWNTO 0);
				 din : IN STD_LOGIC_VECTOR(width-1 DOWNTO 0);
				 dout : OUT STD_LOGIC_VECTOR(width-1 DOWNTO 0)
			  );
end Bram;

architecture Bram_body of Bram is
  type mem_type is array (0 to (2**adr_width)-1) of std_logic_vector(width-1 downto 0);
  signal mem : mem_type := (others => (others => '0'));

	attribute ram_style: string;
	attribute ram_style of mem : signal is "block";
begin

	--
	-- Read side
	--
	process(clk)
	begin
		if rising_edge(clk) then
			if reset = '1' then
				dout <= (others => '0');
			else
				dout <= mem(conv_integer(adr_r));
			end if;
		end if;
	end process;

	--
	-- Write side
	--
	process(clk)
	begin
		if rising_edge(clk) then
			if we = '1' then
				mem(conv_integer(adr_w)) <= din;
			end if;
		end if;
	end process;

end Bram_body;
