library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.OptoUtils.all;

entity LVT is
	generic (
		ports              : natural := 8;
		adr_width          : natural := 10
	);
	Port (
		clk               : IN STD_LOGIC;
		reset          : IN STD_LOGIC;
		enable         : IN STD_LOGIC;
		we             : IN STD_LOGIC_VECTOR(ports-1 DOWNTO 0);
		adr_w          : IN STD_LOGIC_VECTOR(ports * adr_width-1 DOWNTO 0);
		adr_r          : IN STD_LOGIC_VECTOR(ports * adr_width-1 DOWNTO 0);
		r_sel          : OUT STD_LOGIC_VECTOR(ports * log2(ports+1)-1 DOWNTO 0)
	);
end LVT;

architecture LVT_body of LVT is
		constant r_sel_width : natural := log2(ports+1);

		type mem_type is array (0 to (2**adr_width)-1) of std_logic_vector(r_sel_width-1 downto 0);
		signal mem            : mem_type := (others => (others => '0'));

		attribute ram_style: string;
		attribute ram_style of mem : signal is "distributed";
	begin

		--
		-- Read side
		--
		process(clk)
		begin
			if rising_edge(clk) then
				if reset = '1' or enable = '0' then
					r_sel <= (others => '0');
				else
					for I in 0 to ports-1 loop
						r_sel((I+1)*r_sel_width-1 DOWNTO I*r_sel_width) <= mem(conv_integer(adr_r((I+1)*adr_width-1 DOWNTO I*adr_width)));
					end loop;
				end if;
			end if;
		end process;

		--
		-- Write side
		--
		process(clk)
			variable dstAddr : natural;
			variable dstVal  : STD_LOGIC_VECTOR(r_sel_width-1 DOWNTO 0);
		begin
			if rising_edge(clk) then
				if reset = '1' then
					mem <= (others => (others => '0'));
				else
					if enable = '1' then
						for I in 0 to ports-1 loop
							dstVal := (others => '0');
							if we(I) = '1' then
								dstAddr := conv_integer(adr_w((I+1)*adr_width-1 DOWNTO I*adr_width));
								dstVal := std_logic_vector(to_unsigned(I, dstVal'length)+1);
								mem(dstAddr) <= dstVal;
							end if;
						end loop;
					end if;
				end if;
			end if;
		end process;
	end LVT_body;
