----------------------------------------------------------------------------------
-- Company:
-- Engineer:
--
-- Create Date: 11/16/2017 08:12:08 AM
-- Design Name:
-- Module Name: Matcher - Behavioral
-- Project Name:
-- Target Devices:
-- Tool Versions:
-- Description:
--
-- Dependencies:
--
-- Revision:
-- Revision 0.01 - File Created
-- Additional Comments:
--
----------------------------------------------------------------------------------


library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx leaf cells in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity LZ4MatcherUnit is
    Generic(
           hashNators : natural := 8;
           hashNatorsOffset : natural := 8;
           dataWidth : natural := 32;
           addrWidth : natural := 16;
		   hashMemAddrWidth : natural := 8
    );
    Port ( clk : in STD_LOGIC;
           rst : in STD_LOGIC;
           ce : in STD_LOGIC;
           data : in STD_LOGIC_VECTOR ((dataWidth+(hashNators-1)*hashNatorsOffset)-1 downto 0);
           addr : in STD_LOGIC_VECTOR (addrWidth-1 downto 0);
           match : out STD_LOGIC_VECTOR(hashNators-1 downto 0);
           match_addr : out STD_LOGIC_VECTOR (hashNators*addrWidth-1 downto 0);
           match_source_addr : out STD_LOGIC_VECTOR (hashNators*addrWidth-1 downto 0);
           rst_pipe : in STD_LOGIC
                      );
end LZ4MatcherUnit;

architecture Behavioral of LZ4MatcherUnit is
    COMPONENT MatcherCLKUnit
    Port (
        clk_in     : in STD_LOGIC;
        reset      : in STD_LOGIC;
        clk_out    : out STD_LOGIC;
        clk2_out   : out STD_LOGIC;
        locked : out STD_LOGIC);
    END COMPONENT;


	COMPONENT HashNator
    Generic( hashWidth : natural := 14);
    Port ( clk : in STD_LOGIC;
           ce : in STD_LOGIC;
           rst : in STD_LOGIC;
           data : in STD_LOGIC_VECTOR (31 downto 0);
           hash : out STD_LOGIC_VECTOR (hashWidth-1 downto 0));
	END COMPONENT;

	COMPONENT ShiftRegister
    Generic(regDepth : natural := 6;
            dataWidth : natural := 32);
    Port ( clk : in STD_LOGIC;
           rst : in STD_LOGIC;
           ce : in STD_LOGIC;
           din : in STD_LOGIC_VECTOR (dataWidth-1 downto 0);
           dout : out STD_LOGIC_VECTOR (dataWidth-1 downto 0));
	END COMPONENT;

  component MultiPortMem
    generic (
      m_ports     : natural := 8;
      m_width     : natural := 8;
      m_adr_width : natural := 8;
      m_pumping : natural := 1
    );
    port (
      m_clk    : IN  STD_LOGIC;
      m_clk2   : IN  STD_LOGIC;
      m_reset  : IN  STD_LOGIC;
      m_enable : IN  STD_LOGIC;
      m_we     : IN  STD_LOGIC_VECTOR(m_ports-1 DOWNTO 0);
      m_read_valid : OUT STD_LOGIC_VECTOR(m_ports-1 DOWNTO 0);
      m_adr_r    : IN  STD_LOGIC_VECTOR(m_ports*m_adr_width-1 DOWNTO 0);
      m_adr_w    : IN  STD_LOGIC_VECTOR(m_ports*m_adr_width-1 DOWNTO 0);
      m_data_w : IN  STD_LOGIC_VECTOR(m_ports*m_width-1 DOWNTO 0);
      m_data_o : OUT STD_LOGIC_VECTOR(m_ports*m_width-1 DOWNTO 0)
    );
  end component MultiPortMem;



    constant hashWidth : natural := addrWidth;
    constant memDataWidth : natural := (dataWidth+hashWidth);

    signal matcher_rst : STD_LOGIC;
    signal matcher_clk_locked : STD_LOGIC;
    signal matcher_clk : STD_LOGIC;
    signal matcher_clk2 : STD_LOGIC;

    signal hashNator_hashes : STD_LOGIC_VECTOR(hashNators*hashWidth-1 DOWNTO 0);

    signal hashMem_adrIn : STD_LOGIC_VECTOR(hashNators*hashMemAddrWidth-1 DOWNTO 0);
    signal hashMem_adrOut : STD_LOGIC_VECTOR(hashNators*hashMemAddrWidth-1 DOWNTO 0);
    signal hashMem_dataOut : STD_LOGIC_VECTOR(hashNators*memDataWidth-1 DOWNTO 0);
    signal hashMem_dataOut_valid : STD_LOGIC_VECTOR(hashNators-1 DOWNTO 0);
    signal hashMem_dataOut_valid_off : STD_LOGIC_VECTOR(hashNators-1 DOWNTO 0);
    signal hashMem_dataOut_off : STD_LOGIC_VECTOR(hashNators*memDataWidth-1 DOWNTO 0);
    signal hashMem_dataIn : STD_LOGIC_VECTOR(hashNators*memDataWidth-1 DOWNTO 0);
    signal hashMem_dataIn_we : STD_LOGIC_VECTOR(hashNators-1 DOWNTO 0);

    signal hashNator_matches : STD_LOGIC_VECTOR(hashNators-1 DOWNTO 0);
    signal hashNator_sAddresses : STD_LOGIC_VECTOR((hashNators*addrWidth)-1 DOWNTO 0);
    signal hashNator_lAddresses : STD_LOGIC_VECTOR((hashNators*addrWidth)-1 DOWNTO 0);


    signal hashNator_post_matches : STD_LOGIC_VECTOR(hashNators-1 DOWNTO 0);
    signal hashNator_post_sAddresses : STD_LOGIC_VECTOR((hashNators*addrWidth)-1 DOWNTO 0);
    signal hashNator_post_lAddresses : STD_LOGIC_VECTOR((hashNators*addrWidth)-1 DOWNTO 0);

    signal pipelineFull : STD_LOGIC;
    signal pipelineFull_off : STD_LOGIC;

    signal pipelinerst : STD_LOGIC;
begin
   matcher_rst <= rst or not matcher_clk_locked;
   pipelinerst <= matcher_rst or rst_pipe;

  -- CLK_UNIT : MatcherCLKUnit
  -- port map(
  --  clk_in => clk,
  --  reset => rst,
  --  locked => matcher_clk_locked,
  --  clk_out => matcher_clk,
  --  clk2_out => matcher_clk2
  -- );

    matcher_clk_locked <= '1';
    matcher_clk2 <= clk;
    matcher_clk <= clk;

   PIPE_ST_REG : ShiftRegister
               generic map(
                 regDepth => 8,
                 dataWidth => 1
               )
               port map(
                   clk => matcher_clk,
                   rst => pipelinerst,
                   ce => ce,
                   din => "1",
                   dout(0) => pipelineFull);
    hashMem_dataIn_we <= (others => pipelineFull);

    PIPE_POST_ST_REG : ShiftRegister
                generic map(
                  regDepth => 1,
                  dataWidth => 1
                )
                port map(
                    clk => matcher_clk,
                    rst => pipelinerst,
                    ce => ce,
                    din(0) => pipelineFull,
                    dout(0) => pipelineFull_off);

  GEN_HASHNATORS: for I in 0 to hashNators-1 generate
      signal hashNator_data : STD_LOGIC_VECTOR(dataWidth -1 DOWNTO 0);
      signal hashNator_hash : STD_LOGIC_VECTOR(hashWidth -1 DOWNTO 0);
  begin
        hashNator_data <= data(((I*hashNatorsOffset)+32)-1 DOWNTO (I*hashNatorsOffset));
        HASHNATORX : HashNator
        generic map(
          hashWidth => hashWidth
        )
        port map(
            clk => matcher_clk,
            ce => ce,
            rst => pipelinerst,
            data => hashNator_data,
            hash => hashNator_hash
        );
        hashNator_hashes((I+1)*hashWidth-1 DOWNTO I*hashWidth) <= hashNator_hash;
  end generate GEN_HASHNATORS;

  PIPEOFF_ADDR_IN : ShiftRegister
        generic map(
          regDepth => 1,
          dataWidth => hashNators*hashMemAddrWidth
        )
        port map(
          clk => matcher_clk,
          rst => pipelinerst,
          ce => ce,
          din => hashMem_adrOut,
          dout => hashMem_adrIn
        );


--  hashMem_adrIn <= hashNator_hashes;
  GEN_MEMADDR: process(hashNator_hashes)
        variable hashNator_addr : STD_LOGIC_VECTOR(addrWidth -1 DOWNTO 0);
  begin
	for I in 0 to hashNators-1 loop
		hashNator_addr := hashNator_hashes((I+1)*addrWidth-1 DOWNTO (I)*addrWidth);
		hashMem_adrOut((I+1)*hashMemAddrWidth-1 DOWNTO (I)*hashMemAddrWidth) <=
		hashNator_addr(addrWidth-1 DOWNTO addrWidth-hashMemAddrWidth);
	end loop;
  end process;

  GEN_MEMLOGIC: for I in 0 to hashNators-1 generate
        signal hashNator_data : STD_LOGIC_VECTOR(dataWidth -1 DOWNTO 0);
        signal hashNator_addr : STD_LOGIC_VECTOR(addrWidth -1 DOWNTO 0);

        signal hashNator_data_off : STD_LOGIC_VECTOR(dataWidth -1 DOWNTO 0);
        signal hashNator_data_cmp_off : STD_LOGIC_VECTOR(dataWidth -1 DOWNTO 0);
        signal hashNator_data_post_off : STD_LOGIC_VECTOR(dataWidth -1 DOWNTO 0);
        signal hashNator_addr_off : STD_LOGIC_VECTOR(addrWidth -1 DOWNTO 0);
        signal hashNator_addr_cmp_off : STD_LOGIC_VECTOR(addrWidth -1 DOWNTO 0);
        signal hashNator_addr_post_off : STD_LOGIC_VECTOR(addrWidth -1 DOWNTO 0);

        signal hashNator_mem_out : STD_LOGIC_VECTOR(memDataWidth -1 DOWNTO 0);
        signal hashNator_out_data : STD_LOGIC_VECTOR(dataWidth -1 DOWNTO 0);
        signal hashNator_out_addr : STD_LOGIC_VECTOR(addrWidth -1 DOWNTO 0);

        signal hashNator_data_in : STD_LOGIC_VECTOR(dataWidth+addrWidth -1 DOWNTO 0);
        signal hashNator_data_in_off : STD_LOGIC_VECTOR(dataWidth+addrWidth -1 DOWNTO 0);
  begin
        hashNator_addr <= std_logic_vector(unsigned(addr) + I);
        hashNator_data <= data(((I*hashNatorsOffset)+32)-1 DOWNTO (I*hashNatorsOffset));

        PIPEOFF_DATA : ShiftRegister
                generic map(
                  regDepth => 6,
                  dataWidth => dataWidth
                )
                port map(
                    clk => matcher_clk,
                    rst => pipelinerst,
                    ce => ce,
                    din => hashNator_data,
                    dout => hashNator_data_off
                );
        PIPEOFF_ADDR : ShiftRegister
                generic map(
                    regDepth => 6,
                    dataWidth => addrWidth
                )
                port map(
                    clk => matcher_clk,
                    rst => pipelinerst,
                    ce => ce,
                    din => hashNator_addr,
                    dout => hashNator_addr_off
                );


         PIPEOFF_CMP_DATA : ShiftRegister
                 generic map(
                    regDepth => 2,
                    dataWidth => dataWidth
                  )
                  port map(
                    clk => matcher_clk,
                    rst => pipelinerst,
                    ce => ce,
                    din => hashNator_data_off,
                    dout => hashNator_data_cmp_off
                  );

         PIPEOFF_CMP_ADDR : ShiftRegister
                   generic map(
                      regDepth => 2,
                      dataWidth => addrWidth
                   )
                   port map(
                      clk => matcher_clk,
                      rst => pipelinerst,
                      ce => ce,
                      din => hashNator_addr_off,
                      dout => hashNator_addr_cmp_off
                   );



         hashNator_data_in <= hashNator_data_off & hashNator_addr_off;
         PIPEOFF_DATA_IN : ShiftRegister
                   generic map(
                      regDepth => 1,
                      dataWidth => dataWidth + addrWidth
                   )
                   port map(
                      clk => matcher_clk,
                      rst => pipelinerst,
                      ce => ce,
                      din => hashNator_data_in,
                      dout => hashNator_data_in_off
                   );


        hashMem_dataIn((I+1)*memDataWidth-1 DOWNTO I*memDataWidth) <= hashNator_data_in_off;
        hashNator_sAddresses((I+1)*addrWidth-1 DOWNTO I*addrWidth) <= hashNator_addr_cmp_off;

        hashNator_mem_out <= hashMem_dataOut_off((I+1)*memDataWidth-1 DOWNTO I*memDataWidth);
        hashNator_out_data <= hashNator_mem_out(memDataWidth-1 DOWNTO hashWidth);
        hashNator_out_addr <= hashNator_mem_out(hashWidth-1 DOWNTO 0);


        MATCH_PROC : process(hashNator_data_cmp_off, hashNator_out_data, hashNator_out_addr) begin
          hashNator_matches(I) <= '0';
          hashNator_lAddresses((I+1)*addrWidth-1 DOWNTO I*addrWidth) <= (others => '0');

          if(hashMem_dataOut_valid_off(I) = '1' and hashNator_data_cmp_off = hashNator_out_data) then
            hashNator_matches(I) <= '1';
            hashNator_lAddresses((I+1)*addrWidth-1 DOWNTO I*addrWidth) <= hashNator_out_addr;
          end if;
        end process;

        PIPEOFF_POST_DATA : ShiftRegister
               generic map(
                  regDepth => 1,
                  dataWidth => dataWidth
                )
                port map(
                  clk => matcher_clk,
                  rst => pipelinerst,
                  ce => ce,
                  din => hashNator_data_cmp_off,
                  dout => hashNator_data_post_off
                );

        PIPEOFF_POST_ADDR : ShiftRegister
                 generic map(
                    regDepth => 1,
                    dataWidth => addrWidth
                 )
                 port map(
                    clk => matcher_clk,
                    rst => pipelinerst,
                    ce => ce,
                    din => hashNator_addr_cmp_off,
                    dout => hashNator_addr_post_off
                 );

        hashNator_post_sAddresses((I+1)*addrWidth-1 DOWNTO I*addrWidth) <= hashNator_addr_cmp_off;
        POST_MATCH_PROC : process(hashNator_data_cmp_off, hashNator_data_post_off, hashNator_addr_post_off) begin
          hashNator_post_matches(I) <= '0';
          hashNator_post_lAddresses((I+1)*addrWidth-1 DOWNTO I*addrWidth) <= (others => '0');

          if(hashNator_data_cmp_off = hashNator_data_post_off) then -- Todo if post initilized
            hashNator_post_matches(I) <= '1';
            hashNator_post_lAddresses((I+1)*addrWidth-1 DOWNTO I*addrWidth) <= hashNator_addr_post_off;
          end if;
        end process; -- end POST_MATCH_PROC
  end generate GEN_MEMLOGIC;

  MATCH_RED_PROC: process(matcher_clk) begin
    if(rising_edge(matcher_clk)) then
      match <= (others => '0');
      match_addr <= (others => '0');
      match_source_addr <= (others => '0');
      if(pipelinerst = '0' and ce = '1' and pipelineFull = '1') then
        -- Priotita u nizsiho portu --
        for I in 0 to hashNators-1 loop
          if(hashNator_post_matches(I) = '1' and pipelineFull_off = '1') then
            match(I) <= '1';
            match_addr((I+1)*addrWidth-1 DOWNTO I*addrWidth) <= hashNator_post_lAddresses((I+1)*addrWidth-1 DOWNTO I*addrWidth);
            match_source_addr((I+1)*addrWidth-1 DOWNTO I*addrWidth) <= hashNator_post_sAddresses((I+1)*addrWidth-1 DOWNTO I*addrWidth);
          elsif(hashNator_matches(I) = '1' and pipelineFull = '1') then
            match(I) <= '1';
            match_addr((I+1)*addrWidth-1 DOWNTO I*addrWidth) <= hashNator_lAddresses((I+1)*addrWidth-1 DOWNTO I*addrWidth);
            match_source_addr((I+1)*addrWidth-1 DOWNTO I*addrWidth) <= hashNator_sAddresses((I+1)*addrWidth-1 DOWNTO I*addrWidth);
          end if;
        end loop;
      end if;
    end if;
  end process;

  HASHMEM: MultiPortMem
  generic map (
    m_ports => hashNators,
    m_width => memDataWidth,
    m_adr_width => hashMemAddrWidth,
    m_pumping => 1
  )
  port map (
    m_clk => matcher_clk,
    m_clk2 => matcher_clk2,
    m_reset => matcher_rst,
    m_enable => ce,
    m_we => hashMem_dataIn_we,
    m_read_valid => hashMem_dataOut_valid,
    m_adr_r => hashMem_adrOut,
    m_adr_w => hashMem_adrIn,
    m_data_w => hashMem_dataIn,
    m_data_o => hashMem_dataOut
  );


  PIPEOFF_DATA_OUT : ShiftRegister
            generic map(
               regDepth => 1,
               dataWidth => hashNators*memDataWidth
            )
            port map(
               clk => matcher_clk,
               rst => pipelinerst,
               ce => ce,
               din => hashMem_dataOut,
               dout => hashMem_dataOut_off
            );

  PIPEOFF_DATA_VALID_OUT : ShiftRegister
            generic map(
               regDepth => 1,
               dataWidth => hashNators
            )
            port map(
               clk => matcher_clk,
               rst => pipelinerst,
               ce => ce,
               din => hashMem_dataOut_valid,
               dout => hashMem_dataOut_valid_off
            );
end Behavioral;
